@@ -1,13 +1,44 @@
 import sys
 import struct
 import socket
+import re
+import json
+# import urllib
 
 if sys.version < '3':
+    import urllib, httplib
+    def urlencode(x):
+        return urllib.urlencode(x)
+    def httprequest(x):
+        try:
+            conn = httplib.HTTPConnection("api.ip2location.com")
+            conn.request("GET", "/v2/?" + x)
+            res = conn.getresponse()
+            # print (res.read())
+            # print (type(res.read()))
+            # return res.read()
+            return json.loads(res.read())
+        except:
+            return None
     def u(x):
         return x.decode('utf-8')
     def b(x):
         return str(x)
 else:
+    import urllib.parse, http.client
+    def urlencode(x):
+        return urllib.parse.urlencode(x)
+    def httprequest(x):
+        try:
+            conn = http.client.HTTPConnection("api.ip2location.com")
+            conn.request("GET", "/v2/?" + x)
+            res = conn.getresponse()
+            # print (res.read())
+            # print (type(res.read()))
+            return json.loads(res.read())
+            # return res.read()
+        except:
+            return None
     def u(x):
         if isinstance(x, bytes):
             return x.decode()
@@ -64,6 +95,9 @@ class IP2LocationRecord:
     def __repr__(self):
         return repr(self.__dict__)
 
+MAX_IPV4_RANGE = 4294967295
+MAX_IPV6_RANGE = 340282366920938463463374607431768211455
+
 _COUNTRY_POSITION             = (0, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2)
 _REGION_POSITION              = (0, 0, 0, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3)
 _CITY_POSITION                = (0, 0, 0, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4)
@@ -87,10 +121,11 @@ _USAGETYPE_POSITION           = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
 class IP2Location(object):
     ''' IP2Location database '''
 
-    def __init__(self, filename=None):
+    def __init__(self, filename=None,mode='FILE_IO'):
         ''' Creates a database object and opens a file if filename is given
             
         '''
+        self.mode = mode
         if filename:
             self.open(filename)
 
@@ -107,7 +142,16 @@ class IP2Location(object):
         # Ensure old file is closed before opening a new one
         self.close()
 
-        self._f = open(filename, 'rb')
+        if (self.mode == 'SHARED_MEMORY'):
+            import mmap
+            db1 = open(filename, 'r+b')
+            self._f = mmap.mmap(db1.fileno(), 0)
+            db1.close()
+            del db1
+        elif (self.mode == 'FILE_IO'):
+            self._f = open(filename, 'rb')
+        else:
+            raise ValueError("Invalid mode. Please enter either FILE_IO or SHARED_MEMORY.")
         self._dbtype = struct.unpack('B', self._f.read(1))[0]
         self._dbcolumn = struct.unpack('B', self._f.read(1))[0]
         self._dbyear = struct.unpack('B', self._f.read(1))[0]
@@ -120,6 +164,7 @@ class IP2Location(object):
         self._ipv4indexbaseaddr = struct.unpack('<I', self._f.read(4))[0]
         self._ipv6indexbaseaddr = struct.unpack('<I', self._f.read(4))[0]
 
+
     def close(self):
         if hasattr(self, '_f'):
             # If there is file close it.
@@ -226,14 +271,19 @@ class IP2Location(object):
             addr: IPv4 or IPv6 address as a string
      
             Returns IP2LocationRecord or None if address not found in file
+			
+			This function will be obselete in future.
         '''
         return self._get_record(addr)
 
     def _reads(self, offset):
         self._f.seek(offset - 1)
         n = struct.unpack('B', self._f.read(1))[0]
-        #return u(self._f.read(n))
-        return self._f.read(n).decode('iso-8859-1').encode('utf-8')
+        # return u(self._f.read(n))
+        if sys.version < '3':
+            return str(self._f.read(n).decode('iso-8859-1').encode('utf-8'))
+        else :
+            return u(self._f.read(n).decode('iso-8859-1').encode('utf-8'))
 
     def _readi(self, offset):
         self._f.seek(offset - 1)
@@ -262,69 +312,81 @@ class IP2Location(object):
         if ipv == 4:
             off = 0
             baseaddr = self._ipv4dbaddr
+            dbcolumn_width = self._dbcolumn * 4 + 4
         elif ipv == 6:
             off = 12
             baseaddr = self._ipv6dbaddr
-
-        rec.ip = self._readips(baseaddr + (mid) * self._dbcolumn * 4, ipv)
+            dbcolumn_width = self._dbcolumn * 4
 
         def calc_off(what, mid):
             return baseaddr + mid * (self._dbcolumn * 4 + off) + off + 4 * (what[self._dbtype]-1)
 
+        if (self.mode == 'SHARED_MEMORY'):
+            # We can directly use slice notation to read content from mmap object. https://docs.python.org/3/library/mmap.html?highlight=mmap#module-mmap
+            raw_positions_row = self._f[ (calc_off(_COUNTRY_POSITION, mid)) - 1 : (calc_off(_COUNTRY_POSITION, mid)) - 1 + dbcolumn_width]
+        else:
+            self._f.seek((calc_off(_COUNTRY_POSITION, mid)) - 1)
+            raw_positions_row = self._f.read(dbcolumn_width)
+
+        if self.original_ip != '':
+            rec.ip = self.original_ip
+        else:
+            rec.ip = self._readips(baseaddr + (mid) * self._dbcolumn * 4, ipv)
+
         if _COUNTRY_POSITION[self._dbtype] != 0:
-            rec.country_short = self._reads(self._readi(calc_off(_COUNTRY_POSITION, mid)) + 1)
-            rec.country_long =  self._reads(self._readi(calc_off(_COUNTRY_POSITION, mid)) + 4)
+            rec.country_short = self._reads(struct.unpack('<I', raw_positions_row[0 : ((_COUNTRY_POSITION[self._dbtype]-1) * 4)])[0] + 1)
+            rec.country_long =  self._reads(struct.unpack('<I', raw_positions_row[0 : ((_COUNTRY_POSITION[self._dbtype]-1) * 4)])[0] + 4)
 
         if _REGION_POSITION[self._dbtype] != 0:
-            rec.region = self._reads(self._readi(calc_off(_REGION_POSITION, mid)) + 1)
+            rec.region = self._reads(struct.unpack('<I', raw_positions_row[((_REGION_POSITION[self._dbtype]-1) * 4 - 4) : ((_REGION_POSITION[self._dbtype]-1) * 4)])[0] + 1)
         if _CITY_POSITION[self._dbtype] != 0:
-            rec.city = self._reads(self._readi(calc_off(_CITY_POSITION, mid)) + 1)
+            rec.city = self._reads(struct.unpack('<I', raw_positions_row[((_CITY_POSITION[self._dbtype]-1) * 4 - 4) : ((_CITY_POSITION[self._dbtype]-1) * 4)])[0] + 1)
         if _ISP_POSITION[self._dbtype] != 0:
-            rec.isp = self._reads(self._readi(calc_off(_ISP_POSITION, mid)) + 1)
+            rec.isp = self._reads(struct.unpack('<I', raw_positions_row[((_ISP_POSITION[self._dbtype]-1) * 4 - 4) : ((_ISP_POSITION[self._dbtype]-1) * 4)])[0] + 1)
 
         if _LATITUDE_POSITION[self._dbtype] != 0:
-            rec.latitude = round(self._readf(calc_off(_LATITUDE_POSITION, mid)), 6)
+            rec.latitude = round(struct.unpack('<f', raw_positions_row[((_LATITUDE_POSITION[self._dbtype]-1) * 4 - 4) : ((_LATITUDE_POSITION[self._dbtype]-1) * 4)])[0], 6)
         if _LONGITUDE_POSITION[self._dbtype] != 0:
-            rec.longitude = round(self._readf(calc_off(_LONGITUDE_POSITION, mid)), 6)
+            rec.longitude = round(struct.unpack('<f', raw_positions_row[((_LONGITUDE_POSITION[self._dbtype]-1) * 4 - 4) : ((_LONGITUDE_POSITION[self._dbtype]-1) * 4)])[0], 6)
 
         if _DOMAIN_POSITION[self._dbtype] != 0:
-            rec.domain = self._reads(self._readi(calc_off(_DOMAIN_POSITION, mid)) + 1)
+            rec.domain = self._reads(struct.unpack('<I', raw_positions_row[((_DOMAIN_POSITION[self._dbtype]-1) * 4 - 4) : ((_DOMAIN_POSITION[self._dbtype]-1) * 4)])[0] + 1)
 
         if _ZIPCODE_POSITION[self._dbtype] != 0:
-            rec.zipcode = self._reads(self._readi(calc_off(_ZIPCODE_POSITION, mid)) + 1)
+            rec.zipcode = self._reads(struct.unpack('<I', raw_positions_row[((_ZIPCODE_POSITION[self._dbtype]-1) * 4 - 4) : ((_ZIPCODE_POSITION[self._dbtype]-1) * 4)])[0] + 1)
 
         if _TIMEZONE_POSITION[self._dbtype] != 0:
-            rec.timezone = self._reads(self._readi(calc_off(_TIMEZONE_POSITION, mid)) + 1)
+            rec.timezone = self._reads(struct.unpack('<I', raw_positions_row[((_TIMEZONE_POSITION[self._dbtype]-1) * 4 - 4) : ((_TIMEZONE_POSITION[self._dbtype]-1) * 4)])[0] + 1)
                 
         if _NETSPEED_POSITION[self._dbtype] != 0:
-            rec.netspeed = self._reads(self._readi(calc_off(_NETSPEED_POSITION, mid)) + 1)
+            rec.netspeed = self._reads(struct.unpack('<I', raw_positions_row[((_NETSPEED_POSITION[self._dbtype]-1) * 4 - 4) : ((_NETSPEED_POSITION[self._dbtype]-1) * 4)])[0] + 1)
 
         if _IDDCODE_POSITION[self._dbtype] != 0:
-            rec.idd_code = self._reads(self._readi(calc_off(_IDDCODE_POSITION, mid)) + 1)
+            rec.idd_code = self._reads(struct.unpack('<I', raw_positions_row[((_IDDCODE_POSITION[self._dbtype]-1) * 4 - 4) : ((_IDDCODE_POSITION[self._dbtype]-1) * 4)])[0] + 1)
 
         if _AREACODE_POSITION[self._dbtype] != 0:
-            rec.area_code = self._reads(self._readi(calc_off(_AREACODE_POSITION, mid)) + 1)
+            rec.area_code = self._reads(struct.unpack('<I', raw_positions_row[((_AREACODE_POSITION[self._dbtype]-1) * 4 - 4) : ((_AREACODE_POSITION[self._dbtype]-1) * 4)])[0] + 1)
 
         if _WEATHERSTATIONCODE_POSITION[self._dbtype] != 0:
-            rec.weather_code = self._reads(self._readi(calc_off(_WEATHERSTATIONCODE_POSITION, mid)) + 1)
+            rec.weather_code = self._reads(struct.unpack('<I', raw_positions_row[((_WEATHERSTATIONCODE_POSITION[self._dbtype]-1) * 4 - 4) : ((_WEATHERSTATIONCODE_POSITION[self._dbtype]-1) * 4)])[0] + 1)
 
         if _WEATHERSTATIONNAME_POSITION[self._dbtype] != 0:
-            rec.weather_name = self._reads(self._readi(calc_off(_WEATHERSTATIONNAME_POSITION, mid)) + 1)
+            rec.weather_name = self._reads(struct.unpack('<I', raw_positions_row[((_WEATHERSTATIONNAME_POSITION[self._dbtype]-1) * 4 - 4) : ((_WEATHERSTATIONNAME_POSITION[self._dbtype]-1) * 4)])[0] + 1)
 
         if _MCC_POSITION[self._dbtype] != 0:
-            rec.mcc = self._reads(self._readi(calc_off(_MCC_POSITION, mid)) + 1)
+            rec.mcc = self._reads(struct.unpack('<I', raw_positions_row[((_MCC_POSITION[self._dbtype]-1) * 4 - 4) : ((_MCC_POSITION[self._dbtype]-1) * 4)])[0] + 1)
 
         if _MNC_POSITION[self._dbtype] != 0:
-            rec.mnc = self._reads(self._readi(calc_off(_MNC_POSITION, mid)) + 1)
+            rec.mnc = self._reads(struct.unpack('<I', raw_positions_row[((_MNC_POSITION[self._dbtype]-1) * 4 - 4) : ((_MNC_POSITION[self._dbtype]-1) * 4)])[0] + 1)
 
         if _MOBILEBRAND_POSITION[self._dbtype] != 0:
-            rec.mobile_brand = self._reads(self._readi(calc_off(_MOBILEBRAND_POSITION, mid)) + 1)
+            rec.mobile_brand = self._reads(struct.unpack('<I', raw_positions_row[((_MOBILEBRAND_POSITION[self._dbtype]-1) * 4 - 4) : ((_MOBILEBRAND_POSITION[self._dbtype]-1) * 4)])[0] + 1)
                 
         if _ELEVATION_POSITION[self._dbtype] != 0:
-            rec.elevation = self._reads(self._readi(calc_off(_ELEVATION_POSITION, mid)) + 1)
+            rec.elevation = self._reads(struct.unpack('<I', raw_positions_row[((_ELEVATION_POSITION[self._dbtype]-1) * 4 - 4) : ((_ELEVATION_POSITION[self._dbtype]-1) * 4)])[0] + 1)
 
         if _USAGETYPE_POSITION[self._dbtype] != 0:
-            rec.usage_type = self._reads(self._readi(calc_off(_USAGETYPE_POSITION, mid)) + 1)
+            rec.usage_type = self._reads(struct.unpack('<I', raw_positions_row[((_USAGETYPE_POSITION[self._dbtype]-1) * 4 - 4) : ((_USAGETYPE_POSITION[self._dbtype]-1) * 4)])[0] + 1)
 
         return rec
 
@@ -343,27 +405,54 @@ class IP2Location(object):
         ''' Parses address and returns IP version. Raises exception on invalid argument '''
         ipv = 0
         try:
-            socket.inet_pton(socket.AF_INET6, addr)
+            # socket.inet_pton(socket.AF_INET6, addr)
+            a, b = struct.unpack('!QQ', socket.inet_pton(socket.AF_INET6, addr))
+            ipnum = (a << 64) | b
             # Convert ::FFFF:x.y.z.y to IPv4
             if addr.lower().startswith('::ffff:'):
                 try:
                     socket.inet_pton(socket.AF_INET, addr)
                     ipv = 4
                 except:
-                    ipv = 6
+                    # reformat ipv4 address in ipv6 
+                    if ((ipnum >= 281470681743360) and (ipnum <= 281474976710655)):
+                        ipv = 4
+                        ipnum = ipnum - 281470681743360
+                    else:
+                        ipv = 6
             else:
-                ipv = 6
+                #reformat 6to4 address to ipv4 address 2002:: to 2002:FFFF:FFFF:FFFF:FFFF:FFFF:FFFF:FFFF
+                if ((ipnum >= 42545680458834377588178886921629466624) and (ipnum <= 42550872755692912415807417417958686719)):
+                    ipv = 4
+                    ipnum = ipnum >> 80
+                    ipnum = ipnum % 4294967296
+                #reformat Teredo address to ipv4 address 2001:0000:: to 2001:0000:FFFF:FFFF:FFFF:FFFF:FFFF:FFFF:
+                elif ((ipnum >= 42540488161975842760550356425300246528) and (ipnum <= 42540488241204005274814694018844196863)):
+                    ipv = 4
+                    ipnum = ~ ipnum
+                    ipnum = ipnum % 4294967296
+                else:
+                    ipv = 6
         except:
-            socket.inet_pton(socket.AF_INET, addr)
+            ipnum = struct.unpack('!L', socket.inet_pton(socket.AF_INET, addr))[0]
+            # socket.inet_pton(socket.AF_INET, addr)
             ipv = 4
-        return ipv
+        return ipv, ipnum
         
     def _get_record(self, ip):
 
+        # global original_ip
+        self.original_ip = ip
         low = 0
-        ipv = self._parse_addr(ip) 
+        # ipv = self._parse_addr(ip) 
+        ipv = self._parse_addr(ip)[0] 
+        ipnum = self._parse_addr(ip)[1] 
         if ipv == 4:
-            ipno = struct.unpack('!L', socket.inet_pton(socket.AF_INET, ip))[0]
+            # ipno = struct.unpack('!L', socket.inet_pton(socket.AF_INET, ip))[0]
+            if (ipnum == MAX_IPV4_RANGE):
+                ipno = ipnum - 1
+            else:
+                ipno = ipnum
             off = 0
             baseaddr = self._ipv4dbaddr
             high = self._ipv4dbcount
@@ -375,8 +464,12 @@ class IP2Location(object):
         elif ipv == 6:
             if self._ipv6dbcount == 0:
                 raise ValueError('Please use IPv6 BIN file for IPv6 Address.')
-            a, b = struct.unpack('!QQ', socket.inet_pton(socket.AF_INET6, ip))
-            ipno = (a << 64) | b
+            # a, b = struct.unpack('!QQ', socket.inet_pton(socket.AF_INET6, ip))
+            # ipno = (a << 64) | b
+            if (ipnum == MAX_IPV6_RANGE):
+                ipno = ipnum - 1
+            else:
+                ipno = ipnum
             off = 12
             baseaddr = self._ipv6dbaddr
             high = self._ipv6dbcount
@@ -397,3 +490,35 @@ class IP2Location(object):
                     high = mid - 1
                 else:
                     low = mid + 1
+
+class IP2LocationWebService(object):
+    ''' IP2Location web service '''
+    def __init__(self,apikey,package):
+        if ((re.match(r"^[0-9A-Z]{10}$", apikey) == None) and (apikey != 'demo')):
+            raise ValueError("Please provide a valid IP2Location web service API key.")
+        if (re.match(r"^WS[0-9]+$", package) == None):
+            package = 'WS1'
+        self.apikey = apikey
+        self.package = package
+    
+    def lookup(self,ip,addons=[],language='en'):
+        
+        parameters = urlencode((("key", self.apikey), ("ip", ip), ("package", self.package), ("addon", ','.join(addons)), ("lang", language)))
+        response = httprequest(parameters)
+        if (response == None):
+            return False
+        if ('response' in response):
+            raise IP2LocationAPIError(response['response'])
+        return response
+
+    def getcredit(self):
+        parameters = urlencode((("key", self.apikey), ("check", True)))
+        response = httprequest(parameters)
+        if (response == None):
+            return 0
+        if ('response' in response is False):
+            return 0
+        return response['response']
+  
+class IP2LocationAPIError(Exception):
+    """Raise for IP2Location API Error Message"""
\ No newline at end of file
